@@ -316,6 +316,83 @@ class MarketData:
         if os.path.exists(filename):
             return filename
 
+    def standardize_s2f_ratio(self, df):
+        full_mapping = dict(
+            zip(
+                ['t', 'o.daysTillHalving', 'o.ratio'],
+                [C.TIME, C.HALVING, C.RATIO]
+            )
+        )
+        filename = self.finder.get_s2f_path()
+        df = self.standardize(
+            df,
+            full_mapping,
+            filename,
+            [C.TIME, C.HALVING, C.RATIO],
+            0
+        )
+        return df[{C.TIME, C.HALVING, C.RATIO}.intersection(df.columns)]
+
+    def get_s2f_ratio(self, timeframe='max'):
+        # given a symbol, return a cached dataframe
+        df = self.reader.load_csv(
+            self.finder.get_s2f_path())
+        filtered = self.reader.data_in_timeframe(df, C.TIME, timeframe)[
+            [C.TIME, C.HALVING, C.RATIO]]
+        return filtered
+
+    def standardize_s2f_deflection(self, df):
+        full_mapping = dict(
+            zip(
+                ['t', 'v'],
+                [C.TIME, C.VAL]
+            )
+        )
+        filename = self.finder.get_s2f_path()
+        df = self.standardize(
+            df,
+            full_mapping,
+            filename,
+            [C.TIME, C.VAL],
+            1
+        )
+        return df[{C.TIME, C.VAL}.intersection(df.columns)]
+
+    def get_s2f_deflection(self, timeframe='max'):
+        # given a symbol, return a cached dataframe
+        df = self.reader.load_csv(
+            self.finder.get_s2f_path())
+        filtered = self.reader.data_in_timeframe(df, C.TIME, timeframe)[
+            [C.TIME, C.VAL]]
+        return filtered
+
+    def save_s2f(self, **kwargs):
+        # # given a symbol, save its s2f data
+        filename = self.finder.get_s2f_path()
+
+        if os.path.exists(filename):
+            os.remove(filename)
+
+        rat_df = self.reader.update_df(
+            filename, self.get_s2f_ratio(**kwargs), C.TIME, C.DATE_FMT)
+        rat_df = rat_df[{C.TIME, C.HALVING,
+                         C.RATIO}.intersection(rat_df.columns)]
+
+        def_df = self.reader.update_df(
+            filename, self.get_s2f_deflection(**kwargs), C.TIME, C.DATE_FMT)
+        def_df = def_df[{C.TIME, C.VAL}.intersection(def_df.columns)]
+
+        if rat_df.empty and not def_df.empty:
+            df = def_df
+        elif not rat_df.empty and def_df.empty:
+            df = rat_df
+        elif not rat_df.empty and not def_df.empty:
+            df = rat_df.merge(def_df, how="outer", on=C.TIME)
+        else:
+            return
+        self.writer.update_csv(filename, df)
+        if os.path.exists(filename):
+            return filename
     # def handle_request(self, url, err_msg):
 
 
@@ -518,7 +595,9 @@ class Polygon(MarketData):
     def obey_free_limit(self):
         if self.free and hasattr(self, 'last_api_call_time'):
             time_since_last_call = time() - self.last_api_call_time
-            sleep(C.POLY_FREE_DELAY - time_since_last_call)
+            delay = C.POLY_FREE_DELAY - time_since_last_call
+            if delay > 0:
+                sleep(delay)
 
     def log_api_call_time(self):
         self.last_api_call_time = time()
@@ -526,8 +605,12 @@ class Polygon(MarketData):
     def get_dividends(self, **kwargs):
         def _get_dividends(symbol, timeframe='max'):
             self.obey_free_limit()
-            response = self.client.reference_stock_dividends(symbol)
-            self.log_api_call_time()
+            try:
+                response = self.client.reference_stock_dividends(symbol)
+            except Exception as e:
+                raise e
+            finally:
+                self.log_api_call_time()
             raw = pd.DataFrame(response.results)
             df = self.standardize_dividends(symbol, raw)
             return self.reader.data_in_timeframe(df, C.EX, timeframe)
@@ -536,8 +619,12 @@ class Polygon(MarketData):
     def get_splits(self, **kwargs):
         def _get_splits(symbol, timeframe='max'):
             self.obey_free_limit()
-            response = self.client.reference_stock_splits(symbol)
-            self.log_api_call_time()
+            try:
+                response = self.client.reference_stock_splits(symbol)
+            except Exception as e:
+                raise e
+            finally:
+                self.log_api_call_time()
             raw = pd.DataFrame(response.results)
             df = self.standardize_splits(symbol, raw)
             return self.reader.data_in_timeframe(df, C.EX, timeframe)
@@ -549,11 +636,15 @@ class Polygon(MarketData):
             formatted_start, formatted_end = self.traveller.convert_dates(
                 timeframe)
             self.obey_free_limit()
-            response = self.client.stocks_equities_aggregates(
-                symbol, 1, 'day',
-                from_=formatted_start, to=formatted_end, unadjusted=False
-            )
-            self.log_api_call_time()
+            try:
+                response = self.client.stocks_equities_aggregates(
+                    symbol, 1, 'day',
+                    from_=formatted_start, to=formatted_end, unadjusted=False
+                )
+            except Exception as e:
+                raise e
+            finally:
+                self.log_api_call_time()
             raw = response.results
             columns = {'t': 'date', 'o': 'open', 'h': 'high',
                        'l': 'low', 'c': 'close', 'v': 'volume',
@@ -583,11 +674,15 @@ class Polygon(MarketData):
 
             for idx, date in enumerate(dates):
                 self.obey_free_limit()
-                response = self.client.stocks_equities_aggregates(
-                    symbol, min, 'minute', from_=date, to=date,
-                    unadjusted=False
-                )
-                self.log_api_call_time()
+                try:
+                    response = self.client.stocks_equities_aggregates(
+                        symbol, min, 'minute', from_=date, to=date,
+                        unadjusted=False
+                    )
+                except Exception as e:
+                    raise e
+                finally:
+                    self.log_api_call_time()
 
                 if hasattr(response, 'results'):
                     response = response.results
@@ -753,3 +848,72 @@ class LaborStats(MarketData):
             return self.reader.data_in_timeframe(df, C.TIME, timeframe)
 
         return self.try_again(func=_get_unemployment_rate, **kwargs)
+
+
+class Glassnode(MarketData):
+    def __init__(self):
+        super().__init__()
+        self.base = 'https://api.glassnode.com'
+        self.version = 'v1'
+        self.token = os.environ.get('GLASSNODE')
+        self.provider = 'glassnode'
+
+    def get_s2f_ratio(self, **kwargs):
+        def _get_s2f_ratio(timeframe):
+            parts = [
+                self.base,
+                self.version,
+                'metrics',
+                'indicators',
+                'stock_to_flow_ratio'
+            ]
+            url = '/'.join(parts)
+            empty = pd.DataFrame()
+            response = requests.get(
+                url, params={'a': 'BTC', 'api_key': self.token})
+
+            if response.ok:
+                data = response.json()
+            else:
+                raise Exception(
+                    'Invalid response from Glassnode for S2F Ratio')
+
+            if data == []:
+                return empty
+
+            df = pd.json_normalize(data)
+            df['t'] = pd.to_datetime(df['t'], unit='s')
+            df = self.standardize_s2f_ratio(df)
+            return self.reader.data_in_timeframe(df, C.TIME, timeframe)
+
+        return self.try_again(func=_get_s2f_ratio, **kwargs)
+
+    def get_s2f_deflection(self, **kwargs):
+        def _get_s2f_deflection(timeframe):
+            parts = [
+                self.base,
+                self.version,
+                'metrics',
+                'indicators',
+                'stock_to_flow_deflection'
+            ]
+            url = '/'.join(parts)
+            empty = pd.DataFrame()
+            response = requests.get(
+                url, params={'a': 'BTC', 'api_key': self.token})
+
+            if response.ok:
+                data = response.json()
+            else:
+                raise Exception(
+                    'Invalid response from Glassnode for S2F Deflection')
+
+            if data == []:
+                return empty
+
+            df = pd.DataFrame(data)
+            df['t'] = pd.to_datetime(df['t'], unit='s')
+            df = self.standardize_s2f_deflection(df)
+            return self.reader.data_in_timeframe(df, C.TIME, timeframe)
+
+        return self.try_again(func=_get_s2f_deflection, **kwargs)
